/*
******************************************************************
udp_broadcast_fw ; Forwards UDP broadcast packets to all local 
	interfaces as though they originated from sender
	
Copyright (C) 2002  Nathan O'Sullivan

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.
******************************************************************

Thanks:

Arny <cs6171@scitsc.wlv.ac.uk> 
- public domain UDP spoofing code
http://www.netfor2.com/ip.htm
- IP/UDP packet formatting info

*/

#define MAXIFS	256
#define DPRINT  /* printf */
#define IPHEADER_LEN 20
#define UDPHEADER_LEN 8
#define HEADER_LEN (IPHEADER_LEN + UDPHEADER_LEN)
#define DEBUG
#define IP_HDRINCL
 
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in_systm.h>
#include <netinet/in.h>
#include <netinet/ip.h>
#include <netinet/udp.h>
#include <errno.h>
#include <string.h>
#include <arpa/inet.h>
#include <stdio.h>
#include <linux/if.h>
#include <sys/ioctl.h>

main(int argc,char **argv)
{
	/* We use two sockets - one for receiving broadcast packets (type UDP), and
	   one for spoofing them (type RAW) */
	int fd,rcv;
	int buf_size;
	
	/* Structure holds info on local interfaces */
	struct ifconf ifc; int numifs;
	struct ifreq* reqbuf;
		
	/* list of addresses on local machine */
	struct sockaddr_in addr[MAXIFS], toaddr;
	int num_addr = 0;

	/* Address broadcast packet was sent from */
	struct sockaddr_in rcv_addr;
	int err;
	
	int x=1, len;
	
	u_char gram[4096]=
	{
		0x45,	0x00,	0x00,	0x26,
		0x12,	0x34,	0x00,	0x00,
		0xFF,	0x11,	0,	0,
		0,	0,	0,	0,
		0,	0,	0,	0,
		0,	0,	0,	0,
		0x00,	0x12,	0x00,	0x00,
		'1','2','3','4','5','6','7','8','9','0'
	};

  /* Sunrisechat init section */
  FILE *Vystup; /* Output file for storing packets */
  FILE *Vstup; /* Input file for reading packets */

	char *tttt;
  tttt = (char *)malloc(1000);
	char *buf;
  buf = (char *)malloc(10000);
  int i = 0;
  size_t VstupSize;
  char *gramdata;
  int lSize;
  
	/* Error out if no port specified */	
	if(argc!=2)
	{
		fprintf(stderr,"usage: %s udp-port\n",*argv);
		fprintf(stderr,"This program listens for broadcast packets on the specified udp-port\n"
			"and then forwards them to each IP on the local host.  Packets are sent\n"
			"such that they appear to have come from the original broadcaster\n");
		exit(1);
	};

	/* We need to find out what IP's are bound to this host - set up a temporary socket to do so */
 	if((fd=socket(AF_INET,SOCK_RAW,IPPROTO_RAW)) < 0)
	{
  		perror("socket");
			fprintf(stderr,"You must be root to create a raw socket\n");
  		exit(1);
  	};

	/* Use ioctl call to try and find how many interfaces there are. If it fails, just assume 
		there are lots :) */
	if (ioctl(fd, SIOCGIFCOUNT, (char *)&numifs) < 0) 
	{
		DPRINT("ioctl(SIOCGIFCOUNT) failed\n");
      		numifs = MAXIFS;
	}
	
	/* Allocate a buffer to store interface info in */
	if ((reqbuf = (struct ifreq *)malloc( numifs * sizeof(struct ifreq) )) == NULL)
	{
        	perror("malloc");
        	exit(1);
	}
	ifc.ifc_buf = (caddr_t)&reqbuf[0];
	ifc.ifc_len = numifs * sizeof(struct ifreq);
	
	/* Use ioctl call to fill our buffer with information on this machine's interfaces */
	if (ioctl(fd, SIOCGIFCONF, (char *)&ifc) < 0) {
        	perror("ioctl(SIOCGIFCONF)");
        	exit(1);
	}

	/* Since we cant rely on SIOCGIFCOUNT giving an answer, we iterate through the buffers
	   until we hit an interface with a empty name, marking the end of interfaces */
	for (;*reqbuf[0].ifr_name;reqbuf++)
	{
		/* Request flags for this interface */
		if (ioctl(fd,SIOCGIFFLAGS, (char *)reqbuf) < 0) {
			perror("ioctl(SIOCGIFFLAGS)");
			exit(1);
		}

		/* if the interface is not up or a loopback, ignore it */
		if ((reqbuf->ifr_flags & IFF_UP) == 0 ||
      				(reqbuf->ifr_flags & IFF_LOOPBACK) )
			continue;

		/* Request the IP address for this interface */
  		if (ioctl(fd,SIOCGIFADDR, (char *)reqbuf) < 0) {
      			perror("ioctl(SIOCGIFADDR)");
      			exit(1);
    		}

		/* Save the address for later use */
		bcopy((struct sockaddr_in*)&reqbuf[0].ifr_addr,&addr[num_addr],sizeof(struct sockaddr_in) );
		DPRINT("%s\n",inet_ntoa(addr[num_addr].sin_addr) );
		num_addr++;
	}
	
	/* Free our allocated buffer and close the socket */
	free (ifc.ifc_buf);
	close(fd);

	/* Create our broadcast socket */
	if((rcv=socket(AF_INET,SOCK_DGRAM,IPPROTO_UDP)) < 0)
  	{
  		perror("socket");
  		exit(1);
  	};
  	
	x = 1;
	setsockopt(rcv, SOL_SOCKET, SO_BROADCAST, (char*) &x, sizeof(int));

	/* We bind it to broadcast addr on the given port */
	rcv_addr.sin_family = AF_INET;
	rcv_addr.sin_port = htons(atoi(argv[1]));
	rcv_addr.sin_addr.s_addr = INADDR_BROADCAST;

	if ( bind (rcv, (struct sockaddr *)&rcv_addr, sizeof(struct sockaddr_in) ) < 0 )
	{
		perror("bind");
		fprintf(stderr,"A program is already bound to the broadcast address for the given port\n");
		exit(1);
	}
	
	/* Set up a raw socket for sending our packets through */
	if((fd=socket(AF_INET,SOCK_RAW,IPPROTO_RAW)) < 0)
  	{
  		perror("socket");
  		exit(1);
  	};
       
  /* Povol broadcast */
  buf_size = 1;
  if (setsockopt( fd, SOL_SOCKET, SO_BROADCAST, (char *)&buf_size, sizeof(buf_size) )<0)
	{
  	perror("setsockopt SO_BROADCAST");
		exit(1);
  };

  /* Zaka smrovn */
  buf_size = 1;
  if(setsockopt( fd, SOL_SOCKET, SO_DONTROUTE, (char *)&buf_size, sizeof(buf_size) )<0)
  {
		perror("setsockopt SO_DONTROUTE");
		exit(1);
  };

	/* Set dest port to that provided on command line */
	*(u_short*)(gram+22)=(u_short)htons(atoi(argv[1]));

	/* Enable IP header stuff on the raw socket */
	#ifdef IP_HDRINCL
	x=1;
	if (setsockopt(fd,IPPROTO_IP,IP_HDRINCL,(char*)&x,sizeof(x))<0)
	{
		perror("setsockopt IP_HDRINCL");
		exit(1);
  	};
	#else
	#error IP_HDRINCL support is required
	#endif

 	/* Fork to background */
  #ifndef DEBUG
  if (fork())
    exit(0);

  fclose(stdin);
  fclose(stdout);
  fclose(stderr);
  #endif

	for (;;)
	{
		/* Receive a broadcast packet */
		x = sizeof(struct sockaddr_in);
		len = recvfrom(rcv,gram+ HEADER_LEN,4006 - HEADER_LEN - 1,0,(struct sockaddr *)&rcv_addr,&x);
		if (len <= 0) continue;	/* ignore broken packets */

		gram[HEADER_LEN + len] =0; 
		DPRINT("%s\n",gram+HEADER_LEN);
	
		/* copy sender's details into our datagram as the source addr */	
	  bcopy(&(rcv_addr.sin_addr.s_addr),(gram+12),4);
	  	*(u_short*)(gram+20)=(u_short)rcv_addr.sin_port;
    
		/* set the length of the packet */
		*(u_short*)(gram+24)=htons(8 + len);
		*(u_short*)(gram+2)=htons(28+len);

		/* Iterate through the local addresses and send packet to each one */
		for (x=0;x<num_addr;x++)
		{
	
			/* Set destination addr ip - port is set already*/
			bcopy(&(addr[x].sin_addr.s_addr),(gram+16),4);	
			
			DPRINT("%d - %s:%d",x,inet_ntoa(rcv_addr.sin_addr),ntohs(rcv_addr.sin_port) );
			DPRINT (" => %s:%d", inet_ntoa(addr[x].sin_addr),ntohs(*(u_short*)(gram+22)) );
			
			addr[x].sin_addr.s_addr = addr[x].sin_addr.s_addr | 0xff000000;

      /* Skip source interface */
      if((rcv_addr.sin_addr.s_addr & 0xffffff) == (addr[x].sin_addr.s_addr & 0xffffff)) 
			{
  			DPRINT(" source interface\n");
  			continue;
  		} else 
				
      /* Skip internet interface */
      if(((addr[x].sin_addr.s_addr) & 0xff) != 192) 
			{
  			DPRINT(" internet interface\n");
  			continue;
  		} else DPRINT("\n");

			/* Send the packet */
			if (sendto(fd,&gram,28+len,0,(struct sockaddr*)&addr[x],sizeof(struct sockaddr)) < 0)
				perror("sendto");
   	}
    
    /* Send content to file */
    Vystup = fopen("/tmp/sunrisechatout","a+");
    sprintf(tttt,"%s\n",gram+HEADER_LEN);
    fputs (tttt,Vystup);
    fclose(Vystup);

    /* Read content from file */
    Vstup = fopen("/tmp/sunrisechatin","r");
    if (Vstup!=NULL) 
    {  
      fseek (Vstup , 0 , SEEK_END);
      lSize = ftell (Vstup);
      rewind (Vstup);

      VstupSize = fread(buf,1,lSize,Vstup);
      fclose(Vstup);
      remove("/tmp/sunrisechatin");
//      printf("Size: %d",VstupSize);
  //    printf("1\n");
      *(buf+8999) = '\0';
//      DPRINT("%s\n",&buf);
      gramdata = gram + HEADER_LEN;
      len = 0;
      for(i=0;i<VstupSize;i++)
      {
        *(gramdata) = *(buf+i);
        if(*(buf+i) == 10) 
        {
          *(gramdata+1) = 13;
          *(gramdata+2) = 0x0;
          len = len + 2;          

		gram[HEADER_LEN + len] = 0; 
		DPRINT("%d %s\n",len,gram+HEADER_LEN);

		/* copy sender's details into our datagram as the source addr */	
		rcv_addr.sin_addr.s_addr = INADDR_LOOPBACK;
    bcopy(&(rcv_addr.sin_addr.s_addr),(gram+12),4);
	  *(u_short*)(gram+20)=(u_short)rcv_addr.sin_port;
    
		/* set the length of the packet */
		*(u_short*)(gram+24)=htons(8 + len);
		*(u_short*)(gram+2)=htons(28+len);

		/* Iterate through the local addresses and send packet to each one */
		for (x=0;x<num_addr;x++)
		{
	
			/* Set destination addr ip - port is set already*/
			bcopy(&(addr[x].sin_addr.s_addr),(gram+16),4);	
			
			DPRINT("%d - %s:%d",x,inet_ntoa(rcv_addr.sin_addr),ntohs(rcv_addr.sin_port) );
			DPRINT (" => %s:%d\n", inet_ntoa(addr[x].sin_addr),ntohs(*(u_short*)(gram+22)) );
			
			addr[x].sin_addr.s_addr = addr[x].sin_addr.s_addr | 0xff000000;
				
			/* Send the packet */
			if (sendto(fd,&gram,28+len,0,(struct sockaddr*)&addr[x],sizeof(struct sockaddr)) < 0)
				perror("sendto");
   	}
         /* Send content to file */
         Vystup = fopen("/tmp/sunrisechatout","a+");
         sprintf(tttt,"%s\n",gram+HEADER_LEN);
         fputs (tttt,Vystup);
         fclose(Vystup);
                  
          gramdata = gram + HEADER_LEN;
          len = 0;
        } else 
        {
          gramdata++;
          len++;
        }
      }
    }
	}
}



